/*
 *   include/asm-i386/adeos.h
 *
 *   Copyright (C) 2002 Philippe Gerum.
 *
 *   This program is free software; you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as published by
 *   the Free Software Foundation, Inc., 675 Mass Ave, Cambridge MA 02139,
 *   USA; either version 2 of the License, or (at your option) any later
 *   version.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this program; if not, write to the Free Software
 *   Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */

#ifndef __I386_ADEOS_H
#define __I386_ADEOS_H

#include <irq_vectors.h>
#include <asm/ptrace.h>
#include <linux/cpumask.h>
#include <linux/list.h>
#include <linux/threads.h>

#define ADEOS_ARCH_STRING   "r12/x86"
#define ADEOS_MAJOR_NUMBER  12
#define ADEOS_MINOR_NUMBER  255

extern int adp_pipelined;

#ifdef CONFIG_SMP

#include <asm/fixmap.h>
#include <asm/mpspec.h>
#include <mach_apicdef.h>

#define ADEOS_NR_CPUS          NR_CPUS
#define ADEOS_CRITICAL_VECTOR  0xf9 /* Used by adeos_critical_enter/exit() */
#define ADEOS_CRITICAL_IPI     (ADEOS_CRITICAL_VECTOR - FIRST_EXTERNAL_VECTOR)

static __inline int adeos_smp_apic_id(void) {
    return GET_APIC_ID(*(unsigned long *)(APIC_BASE+APIC_ID));
}

extern volatile u8 __adeos_apicid_2_cpuid[];

int __adeos_hw_cpuid(void);

cpumask_t __adeos_set_irq_affinity(unsigned irq,
				   cpumask_t cpumask);

#define adeos_processor_id()   __adeos_hw_cpuid()

#define adeos_declare_cpuid    int cpuid
#define adeos_load_cpuid()     do { \
                                  (cpuid) = adeos_processor_id();	\
                               } while(0)
#define adeos_lock_cpu(flags)  do { \
                                  adeos_hw_local_irq_save(flags); \
                                  (cpuid) = adeos_processor_id(); \
                               } while(0)
#define adeos_unlock_cpu(flags) adeos_hw_local_irq_restore(flags)
#define adeos_get_cpu(flags)    adeos_lock_cpu(flags)
#define adeos_put_cpu(flags)    adeos_unlock_cpu(flags)
#define adp_current             (adp_cpu_current[adeos_processor_id()])

#else  /* !CONFIG_SMP */

#define ADEOS_NR_CPUS          1
#define adeos_processor_id()   0
/* Array references using this index should be optimized out. */
#define adeos_declare_cpuid    const int cpuid = 0
#define adeos_load_cpuid()      /* nop */
#define adeos_lock_cpu(flags)   adeos_hw_local_irq_save(flags)
#define adeos_unlock_cpu(flags) adeos_hw_local_irq_restore(flags)
#define adeos_get_cpu(flags)    do { flags = flags; } while(0)
#define adeos_put_cpu(flags)    /* nop */
#define adp_current             (adp_cpu_current[0])

#endif /* CONFIG_SMP */

 /* IDT fault vectors */
#define ADEOS_NR_FAULTS         32
/* Pseudo-vectors used for kernel events */
#define ADEOS_FIRST_KEVENT      ADEOS_NR_FAULTS
#define ADEOS_SYSCALL_PROLOGUE  (ADEOS_FIRST_KEVENT)
#define ADEOS_SYSCALL_EPILOGUE  (ADEOS_FIRST_KEVENT + 1)
#define ADEOS_SCHEDULE_HEAD     (ADEOS_FIRST_KEVENT + 2)
#define ADEOS_SCHEDULE_TAIL     (ADEOS_FIRST_KEVENT + 3)
#define ADEOS_ENTER_PROCESS     (ADEOS_FIRST_KEVENT + 4)
#define ADEOS_EXIT_PROCESS      (ADEOS_FIRST_KEVENT + 5)
#define ADEOS_SIGNAL_PROCESS    (ADEOS_FIRST_KEVENT + 6)
#define ADEOS_KICK_PROCESS      (ADEOS_FIRST_KEVENT + 7)
#define ADEOS_RENICE_PROCESS    (ADEOS_FIRST_KEVENT + 8)
#define ADEOS_USER_EVENT        (ADEOS_FIRST_KEVENT + 9)
#define ADEOS_LAST_KEVENT       (ADEOS_USER_EVENT)

#define ADEOS_NR_EVENTS         (ADEOS_LAST_KEVENT + 1)

typedef struct adevinfo {

    unsigned domid;
    unsigned event;
    void *evdata;

    volatile int propagate;	/* Private */

} adevinfo_t;

typedef struct adsysinfo {

    int ncpus;			/* Number of CPUs on board */

    u64 cpufreq;		/* CPU frequency (in Hz) */

    /* Arch-dependent block */

    struct {
	unsigned tmirq;		/* Timer tick IRQ */
	u64 tmfreq;		/* Timer frequency */
    } archdep;

} adsysinfo_t;

#ifdef CONFIG_X86_LOCAL_APIC
/* We must cover the whole IRQ space to map the local timer interrupt
   (#207). */
#ifdef CONFIG_PCI_MSI
#define IPIPE_NR_XIRQS NR_IRQS
#else /* CONFIG_PCI_MSI */
#define IPIPE_NR_XIRQS   224
#endif /* CONFIG_PCI_MSI */
/* If the APIC is enabled, then we expose four service vectors in the
   APIC space which are freely available to domains. */
#define ADEOS_SERVICE_VECTOR0  0xf5
#define ADEOS_SERVICE_IPI0     (ADEOS_SERVICE_VECTOR0 - FIRST_EXTERNAL_VECTOR)
#define ADEOS_SERVICE_VECTOR1  0xf6
#define ADEOS_SERVICE_IPI1     (ADEOS_SERVICE_VECTOR1 - FIRST_EXTERNAL_VECTOR)
#define ADEOS_SERVICE_VECTOR2  0xf7
#define ADEOS_SERVICE_IPI2     (ADEOS_SERVICE_VECTOR2 - FIRST_EXTERNAL_VECTOR)
#define ADEOS_SERVICE_VECTOR3  0xf8
#define ADEOS_SERVICE_IPI3     (ADEOS_SERVICE_VECTOR3 - FIRST_EXTERNAL_VECTOR)
#else /* !CONFIG_X86_LOCAL_APIC */
#define IPIPE_NR_XIRQS   NR_IRQS
#endif /* CONFIG_X86_LOCAL_APIC */

/* Number of virtual IRQs */
#define IPIPE_NR_VIRQS   BITS_PER_LONG
/* First virtual IRQ # */
#define IPIPE_VIRQ_BASE  (((IPIPE_NR_XIRQS + BITS_PER_LONG - 1) / BITS_PER_LONG) * BITS_PER_LONG)
/* Total number of IRQ slots */
#define IPIPE_NR_IRQS     (IPIPE_VIRQ_BASE + IPIPE_NR_VIRQS)
/* Number of indirect words needed to map the whole IRQ space. */
#define IPIPE_IRQ_IWORDS  ((IPIPE_NR_IRQS + BITS_PER_LONG - 1) / BITS_PER_LONG)
#define IPIPE_IRQ_IMASK   (BITS_PER_LONG - 1)
#define IPIPE_IRQ_ISHIFT  5	/* 2^5 for 32bits arch. */

#define IPIPE_IRQMASK_ANY   (~0L)
#define IPIPE_IRQMASK_VIRT  (IPIPE_IRQMASK_ANY << (IPIPE_VIRQ_BASE / BITS_PER_LONG))

typedef struct adomain {

    /* -- Section: offset-based references are made on these fields
       from inline assembly code. Please don't move or reorder. */
    void (*dswitch)(void);	/* Domain switch hook */
#ifdef CONFIG_ADEOS_THREADS
    int *esp[ADEOS_NR_CPUS];	/* Domain stack pointers */
#endif /* CONFIG_ADEOS_THREADS */
    /* -- End of section. */

    struct list_head p_link;	/* Link in pipeline */

    struct adcpudata {
	volatile unsigned long status;
	volatile unsigned long irq_pending_hi;
	volatile unsigned long irq_pending_lo[IPIPE_IRQ_IWORDS];
	volatile unsigned irq_hits[IPIPE_NR_IRQS];
#ifdef CONFIG_ADEOS_THREADS
	adevinfo_t event_info;
#endif /* CONFIG_ADEOS_THREADS */
    } cpudata[ADEOS_NR_CPUS];

    struct {
	int (*acknowledge)(unsigned irq);
	void (*handler)(unsigned irq);
	unsigned long control;
    } irqs[IPIPE_NR_IRQS];

    struct {
	void (*handler)(adevinfo_t *evinfo);
    } events[ADEOS_NR_EVENTS];

    struct adomain *m_link;	/* Link in mutex sleep queue */

    unsigned long flags;

    unsigned domid;

    const char *name;

    int priority;

    int ptd_keymax;
    int ptd_keycount;
    unsigned long ptd_keymap;
    void (*ptd_setfun)(int, void *);
    void *(*ptd_getfun)(int);

#ifdef CONFIG_ADEOS_THREADS
    int *estackbase[ADEOS_NR_CPUS];
#endif /* CONFIG_ADEOS_THREADS */

} adomain_t;

/* The following macros must be used hw interrupts off. */

#define __adeos_set_irq_bit(adp,cpuid,irq) \
do { \
    if (!test_bit(IPIPE_LOCK_FLAG,&(adp)->irqs[irq].control)) { \
        __set_bit(irq & IPIPE_IRQ_IMASK,&(adp)->cpudata[cpuid].irq_pending_lo[irq >> IPIPE_IRQ_ISHIFT]); \
        __set_bit(irq >> IPIPE_IRQ_ISHIFT,&(adp)->cpudata[cpuid].irq_pending_hi); \
       } \
} while(0)

#define __adeos_clear_pend(adp,cpuid,irq) \
do { \
    __clear_bit(irq & IPIPE_IRQ_IMASK,&(adp)->cpudata[cpuid].irq_pending_lo[irq >> IPIPE_IRQ_ISHIFT]); \
    if ((adp)->cpudata[cpuid].irq_pending_lo[irq >> IPIPE_IRQ_ISHIFT] == 0) \
        __clear_bit(irq >> IPIPE_IRQ_ISHIFT,&(adp)->cpudata[cpuid].irq_pending_hi); \
} while(0)

#define __adeos_lock_irq(adp,cpuid,irq) \
do { \
    if (!test_and_set_bit(IPIPE_LOCK_FLAG,&(adp)->irqs[irq].control)) \
	__adeos_clear_pend(adp,cpuid,irq); \
} while(0)

#define __adeos_unlock_irq(adp,irq) \
do { \
    int __cpuid, __nr_cpus = num_online_cpus();			       \
    if (test_and_clear_bit(IPIPE_LOCK_FLAG,&(adp)->irqs[irq].control)) \
	for (__cpuid = 0; __cpuid < __nr_cpus; __cpuid++)      \
         if ((adp)->cpudata[__cpuid].irq_hits[irq] > 0) { /* We need atomic ops next. */ \
           set_bit(irq & IPIPE_IRQ_IMASK,&(adp)->cpudata[__cpuid].irq_pending_lo[irq >> IPIPE_IRQ_ISHIFT]); \
           set_bit(irq >> IPIPE_IRQ_ISHIFT,&(adp)->cpudata[__cpuid].irq_pending_hi); \
         } \
} while(0)

#define __adeos_clear_irq(adp,irq) \
do { \
    int __cpuid, __nr_cpus = num_online_cpus(); \
    clear_bit(IPIPE_LOCK_FLAG,&(adp)->irqs[irq].control); \
    for (__cpuid = 0; __cpuid < __nr_cpus; __cpuid++) {	\
       (adp)->cpudata[__cpuid].irq_hits[irq] = 0; \
       __adeos_clear_pend(adp,__cpuid,irq); \
    } \
} while(0)

#define adeos_virtual_irq_p(irq) ((irq) >= IPIPE_VIRQ_BASE && \
				  (irq) < IPIPE_NR_IRQS)

#define adeos_hw_save_flags_and_sti(x)	__asm__ __volatile__("pushfl ; popl %0 ; sti":"=g" (x): /* no input */ :"memory")
#define adeos_hw_cli() 			__asm__ __volatile__("cli": : :"memory")
#define adeos_hw_sti()			__asm__ __volatile__("sti": : :"memory")
#define adeos_hw_local_irq_save(x)    __asm__ __volatile__("pushfl ; popl %0 ; cli":"=g" (x): /* no input */ :"memory")
#define adeos_hw_local_irq_restore(x) __asm__ __volatile__("pushl %0 ; popfl": /* no output */ :"g" (x):"memory", "cc")
#define adeos_hw_local_irq_flags(x)   __asm__ __volatile__("pushfl ; popl %0":"=g" (x): /* no input */)
#define adeos_hw_test_iflag(x)        ((x) & (1<<9))
#define adeos_hw_irqs_disabled()	\
({					\
	unsigned long flags;		\
	adeos_hw_local_irq_flags(flags);	\
	!adeos_hw_test_iflag(flags);	\
})

#define adeos_hw_tsc(t)  __asm__ __volatile__("rdtsc" : "=A" (t))
#define adeos_cpu_freq() ({ unsigned long long __freq = cpu_has_tsc?(1000LL * cpu_khz):CLOCK_TICK_RATE; __freq; })

#ifdef CONFIG_ADEOS_PREEMPT_RT
/* We are over a combo Adeos+PREEMPT_RT _patched_ kernel, but
   CONFIG_PREEMPT_RT is not necessarily enabled; use the raw spinlock
   support for Adeos. */
#define adeos_spin_lock(x)     __raw_spin_lock(x)
#define adeos_spin_unlock(x)   __raw_spin_unlock(x)
#define adeos_spin_trylock(x)  __raw_spin_trylock(x)
#define adeos_write_lock(x)    __raw_write_lock(x)
#define adeos_write_unlock(x)  __raw_write_unlock(x)
#define adeos_write_trylock(x) __raw_write_trylock(x)
#define adeos_read_lock(x)     __raw_read_lock(x)
#define adeos_read_unlock(x)   __raw_read_unlock(x)
#else /* !CONFIG_ADEOS_PREEMPT_RT */
#define adeos_spin_lock(x)     _spin_lock(x)
#define adeos_spin_unlock(x)   _spin_unlock(x)
#define adeos_spin_trylock(x)  _spin_trylock(x)
#define adeos_write_lock(x)    _write_lock(x)
#define adeos_write_unlock(x)  _write_unlock(x)
#define adeos_write_trylock(x) _write_trylock(x)
#define adeos_read_lock(x)     _read_lock(x)
#define adeos_read_unlock(x)   _read_unlock(x)
#define raw_spinlock_t         spinlock_t
#define RAW_SPIN_LOCK_UNLOCKED SPIN_LOCK_UNLOCKED
#define raw_rwlock_t           rwlock_t
#define RAW_RW_LOCK_UNLOCKED   RW_LOCK_UNLOCKED
#endif /* CONFIG_ADEOS_PREEMPT_RT */

#define spin_lock_irqsave_hw(lock,flags)      adeos_spin_lock_irqsave(lock, flags)
#define spin_unlock_irqrestore_hw(lock,flags) adeos_spin_unlock_irqrestore(lock, flags)

#define adeos_spin_lock_irqsave(x,flags)  \
do { \
   adeos_hw_local_irq_save(flags); \
   adeos_spin_lock(x); \
} while (0)

#define adeos_spin_unlock_irqrestore(x,flags)  \
do { \
   adeos_spin_unlock(x); \
   adeos_hw_local_irq_restore(flags); \
} while (0)

#define adeos_spin_lock_disable(x)  \
do { \
   adeos_hw_cli(); \
   adeos_spin_lock(x); \
} while (0)

#define adeos_spin_unlock_enable(x)  \
do { \
   adeos_spin_unlock(x); \
   adeos_hw_sti(); \
} while (0)

#define adeos_read_lock_irqsave(lock, flags) \
do { \
   adeos_hw_local_irq_save(flags); \
   adeos_read_lock(lock); \
} while (0)

#define adeos_read_unlock_irqrestore(lock, flags) \
do { \
   adeos_read_unlock(lock); \
   adeos_hw_local_irq_restore(flags); \
} while (0)

#define adeos_write_lock_irqsave(lock, flags) \
do { \
   adeos_hw_local_irq_save(flags); \
   adeos_write_lock(lock); \
} while (0)

#define adeos_write_unlock_irqrestore(lock, flags) \
do { \
   adeos_write_unlock(lock); \
   adeos_hw_local_irq_restore(flags); \
} while (0)

#ifndef STR
#define __STR(x) #x
#define STR(x) __STR(x)
#endif

#ifndef SYMBOL_NAME_STR
#define SYMBOL_NAME_STR(X) #X
#endif

/* Private interface -- Internal use only */

struct adattr;

void __adeos_init(void);

void __adeos_init_domain(adomain_t *adp,
			 struct adattr *attr);

void __adeos_cleanup_domain(adomain_t *adp);

void __adeos_send_IPI_mask(cpumask_t mask,
			   int vector);

void __adeos_send_IPI_allbutself(int vector);

#ifdef CONFIG_ADEOS_THREADS

#ifdef CONFIG_SMP

/*
 * __adeos_switch_to() -- Switch domain contexts. The current domain
 * ("out") is switched out while the domain pointed by "in" is
 * switched in. The current cpu identifier which is always known from
 * callers is also passed to save a few cycles.  This code works out
 * the following tasks: - build a resume frame for the suspended
 * domain, - save it's stack pointer, - load the incoming domain's
 * stack pointer, - update the global domain descriptor pointer, -
 * then finally activate the resume frame of the incoming domain.
 *
 * SMP version also provides for safe CPU migration (i.e. the domain
 * may be switched back in on behalf of a different CPU than the one
 * which switched it out).
 *
 * NOTE: dswitch() takes no argument, so there is no specific handling
 * of CONFIG_REGPARM needed. This routine must be called with hw
 * interrupts off.
 */

static inline void __adeos_switch_to (adomain_t *out, adomain_t *in, int cpuid)

{
    __asm__ __volatile__( \
	"pushl %%ebp\n\t" \
	"pushl %%ecx\n\t" \
	"pushl %%ebx\n\t" \
	"pushl %%edi\n\t" \
	"pushl %%esi\n\t" \
	"movl %%eax, %%ecx\n\t" \
	"leal adp_cpu_current(,%%eax,4),%%eax\n\t" \
	"xchg (%%eax), %%edx\n\t" \
	"pushl %%edx\n\t" \
	"pushl $1f\n\t" \
	"sall $2,%%ecx\n\t" \
	"addl $4,%%ecx\n\t" \
	"movl %%esp, (%%ecx,%%edx)\n\t" \
	"movl (%%eax), %%eax\n\t" \
	"movl (%%ecx,%%eax), %%esp\n\t" \
	"ret\n\t" \
	  /* Call domain switch hook (if any) */
"1:      popl %%eax\n\t" \
	"movl (%%eax), %%eax\n\t" \
	"testl %%eax,%%eax\n\t" \
	"je 2f\n\t" \
	"call *%%eax\n\t" \
	  /* Domain resume point */
"2:      popl %%esi\n\t" \
	"popl %%edi\n\t" \
	"popl %%ebx\n\t" \
	"popl %%ecx\n\t" \
	"popl %%ebp\n\t" \
	: /* no output */ \
	: "a" (cpuid), "d" (in) \
	: "memory", "cc");

    barrier();
}

#else /* !CONFIG_SMP */

static inline void __adeos_switch_to (adomain_t *out, adomain_t *in, int cpuid)

{
    __asm__ __volatile__( \
	"pushl %%ebp\n\t" \
	"pushl %%edx\n\t" \
	"pushl %%ecx\n\t" \
	"pushl %%ebx\n\t" \
	"pushl %%edi\n\t" \
	"pushl %%esi\n\t" \
	"movl "SYMBOL_NAME_STR(adp_cpu_current)", %%edx\n\t" \
	"pushl %%edx\n\t" \
	"pushl $1f\n\t" \
	"movl %%esp, 4(%%edx)\n\t" \
	"movl 4(%%eax), %%esp\n\t" \
	"movl %%eax, "SYMBOL_NAME_STR(adp_cpu_current)"\n\t" \
	"ret\n\t" \
	  /* Call domain switch hook (if any) */
"1:      popl %%eax\n\t" \
	"movl (%%eax),%%eax\n\t" \
	"testl %%eax,%%eax\n\t" \
	"je 2f\n\t" \
	"call *%%eax\n\t" \
	  /* Domain resume point */
"2:      popl %%esi\n\t" \
	"popl %%edi\n\t" \
	"popl %%ebx\n\t" \
	"popl %%ecx\n\t" \
	"popl %%edx\n\t" \
	"popl %%ebp\n\t" \
	: /* no output */ \
        : "a" (in));
}

#endif /* CONFIG_SMP */

#endif /* CONFIG_ADEOS_THREADS */

#define __adeos_check_platform() do { } while(0)

#define __adeos_init_platform() do { } while(0)

void __adeos_enable_pipeline(void);

void __adeos_disable_pipeline(void);

void __adeos_init_stage(adomain_t *adp);

void fastcall __adeos_sync_stage(unsigned long syncmask);

int __adeos_ack_system_irq(unsigned irq);

int __adeos_handle_irq(struct pt_regs regs);

extern struct pt_regs __adeos_tick_regs[];

extern int __adeos_tick_irq;

#ifdef CONFIG_X86_LOCAL_APIC
extern int __adeos_apic_mapped;
#endif /* CONFIG_X86_LOCAL_APIC */

#endif /* !__I386_ADEOS_H */
